
module DDC where

import System.Random
import Control.Monad.State
import Data.List

-- BASIC STUFF HERE

data Command = C LampID Color | W Int
  deriving Show

renderCommand (W n) = "W "++(show n)++"\n"
renderCommand (C id col) = "C "++(renderID id)++" "++(renderColor col)++"\n"

data LampID = SpecificLamp BulbID | AllLamp | Random | SpecificLetter Letter
  deriving Show

data Letter = LeftLetter | MiddleLetter | RightLetter deriving (Show, Eq)
instance Random Letter where
  random org = let
      l = [LeftLetter, MiddleLetter, RightLetter] !! n
      (n, nrg) = randomR (0,2) org
    in (l,nrg)

data BulbID = BulbID Char deriving Show

renderID (SpecificLetter LeftLetter) = "1"
renderID (SpecificLetter MiddleLetter) = "2"
renderID (SpecificLetter RightLetter) = "3"

renderID (SpecificLamp (BulbID c)) = [c]
renderID AllLamp = "ALL"

data Color = Color Int Int Int -- TODO also random here and other opts
  deriving Show

instance Random Color where
  random rng1 = let
      (r, rng2) = randomR (0,255) rng1 
      (g, rng3) = randomR (0,255) rng2
      (b, rng4) = randomR (0,255) rng3
    in (Color r g b, rng4)

renderColor (Color r g b) = "#"++(padshow r)++(padshow g)++(padshow b)

padshow n = let 
    digits="0123456789ABCDEF"
    lsf = digits !! (n `rem` 16)
    msf = digits !! (n `quot` 16)
  in [msf,lsf]


black = Color 0 0 0
red = Color 255 0 0
green = Color 0 255 0
yellow = Color 255 255 0
blue = Color 0 0 255
magenta = Color 255 0 255
cyan = Color 0 255 255
white = Color 255 255 255

-- SpecificLamp is A-O, SpecificLetter is 1-3

-- LETTER GEOMETRY (same for all three)
--    0  1
-- 2  D3 E4
-- 1  C2
-- 0  B1 A0

-- map each lamp to its x,y co-ord in the letter

letterComponents = [ (0, 1, 0),
                     (1, 0, 0),
                     (2, 0, 1),
                     (3, 0, 2),
                     (4, 1, 2) ]


allLamps1 = [ (bulbmajor, l, x, y) | bulbmajor <- [0,1,2], (l,x,y) <- letterComponents]

allLamps :: [(BulbID, Letter, Int, Int)]
allLamps = map (\(k, l, x, y) -> (BulbID ("ABCDEFGHIJKLMNO"!!(l + 5*k)),[LeftLetter, MiddleLetter, RightLetter]!!k,x,y)) allLamps1


-- some general macros

-- run something in the DDC monad

runDDC :: State DDCMonadState () -> IO ()
runDDC r = let
    (commands, rand) = execState r ("", mkStdGen ddcmonadseed)
  in putStrLn commands

ddcmonadseed = 0

-- output commands, ongoing random number
type DDCMonadState = ([Char], StdGen)

rawCommand :: Command -> State DDCMonadState ()
rawCommand cmd = do
  (commands, b) <- get
  put (commands ++ (renderCommand cmd), b)

randomDDC :: Random a => State DDCMonadState a
randomDDC = do
  (cmds, oldrg) <- get
  let (r, newrg) = random oldrg
  put (cmds, newrg)
  return r

randomRangeDDC range = do
  (cmds, oldrg) <- get
  let (r, newrg) = randomR range oldrg
  put (cmds, newrg)
  return r

shuffle l = do
  al <- augmentListWithRandoms l
  let ord = \(l,_) -> \(r,_) -> compare l r
  let bl = sortBy ord al
  let cl = map snd bl
  return cl

augmentListWithRandoms l = do
  if length l == 0 then return [] else
    do
      rest <- augmentListWithRandoms (tail l)
      rnd <- randomRangeDDC (0::Float,100::Float)
      return $ (rnd, head l):rest




wait n  = rawCommand $ W n

colorLetter letter color = rawCommand $ C (SpecificLetter letter) color

colorEverything color = rawCommand $ C AllLamp color
blankEverything = colorEverything black

colorRow r c = forM_ (lampsForRow r)
               (\l-> rawCommand $ C (SpecificLamp l) c)

lampsForRow r = map (\(l,_,_,_) -> l) $ filter (\(_, _, _, y) -> y == r) allLamps

colorColumn column color = forM_ (lampsForColumn column)
                   (\l-> rawCommand $ C (SpecificLamp l) color)

-- column from 0 to 5 -- TODO
lampsForColumn column = 
 let
    vc = validateColumn column
    l2 = [LeftLetter, MiddleLetter, RightLetter] !! (vc `quot` 2)
    c2 = vc `rem` 2
  in map (\(l,_,_,_) -> l) $ filter (\(_, ltr, x, _) -> x == c2 && ltr == l2) allLamps

colorLamp lamp color = rawCommand $ C lamp color

columnToLetterColumnPair :: Int -> (Letter,Int)
columnToLetterColumnPair n = 
 let
    l2 = [LeftLetter, MiddleLetter, RightLetter] !! (n `quot` 2)
    c2 = n `rem` 2
  in (l2,c2)

validateColumn c = if c>=0 && c<=5 then c else error $ "Column out of range: " ++ (show c)
